//------------------------------------------------------------------------------
//
// Copyright (c) 2008-2015 IAR Systems
//
// Licensed under the Apache License, Version 2.0 (the "License")
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// $Revision: 38952 $
//
//------------------------------------------------------------------------------

#include "flash_loader.h"
#include "flash_loader_extra.h"

/**********************************************************/
#define EFM_INTERNAL_PAGE_SIZE (8) // for program
#define EFM_INTERNAL_SECTOR_SIZE (512) // for erase

#define EFM_REG_BASE                    (0x40020000)
#define EFM_CTRL                        (*((volatile uint32_t*) (EFM_REG_BASE + 0x00)))
#define EFM_STAT                        (*((volatile uint32_t*) (EFM_REG_BASE + 0x04)))
#define EFM_CMD                         (*((volatile uint32_t*) (EFM_REG_BASE + 0x08)))
#define EFM_KEY                         (*((volatile uint32_t *)(EFM_REG_BASE + 0x200)))

#define EFM_STAT_DONE_MASK              (0x01U << 0)
#define EFM_STAT_CCIF_MASK              (0x01U << 7)
#define EFM_STAT_ACCERR_MASK            (0x01U << 3)
#define EFM_STAT_UNRECOVERY_ERR_MASK    (0x01U << 2)
#define EFM_STAT_RECOVERY_MASK          (0x01U << 1)
#define EFM_CTRL_WE_MASK                (0x01U << 12)

#define EFM_NOP_CMD                     (0x00U)
#define EFM_PROG_64_CMD                 (0x01U)
#define EFM_PROG_CMD                    (0x02U)
#define EFM_SECTOR_ERASE_CMD            (0x10U)
#define EFM_BLOCK_ERASE_CMD             (0x11U)
#define EFM_ERROR_MASK                  (EFM_STAT_ACCERR_MASK | EFM_STAT_UNRECOVERY_ERR_MASK | EFM_STAT_RECOVERY_MASK)
#define EFM_UNLOCK_KEY                  (0xfd9573f5)

#define EFM_CTRL_PRESCALER_MASK                            (0x1F0000U)
#define EFM_CTRL_PRESCALER_SHIFT                           (16U)
#define EFM_CTRL_PRESCALER(x)                              (((uint32_t)(((uint32_t)(x)) << EFM_CTRL_PRESCALER_SHIFT)) & EFM_CTRL_PRESCALER_MASK)

#define WDG_REG_BASE                    (0x40052000)
#define WDG_UNLOCK                      (*((volatile uint32_t *)(WDG_REG_BASE + 0x00)))
#define WDG_CONTROL                     (*((volatile uint32_t *)(WDG_REG_BASE + 0x04)))
#define WDG_TO                          (*((volatile uint32_t *)(WDG_REG_BASE + 0x0C)))

#define SCU_REG_BASE                    (0x40064000)
#define SCU_CLKS                        (*((volatile uint32_t *)(SCU_REG_BASE + 0x00)))
#define SCU_STS                         (*((volatile uint32_t *)(SCU_REG_BASE + 0x04)))
#define SCU_OSC_CTRL                    (*((volatile uint32_t *)(SCU_REG_BASE + 0x08)))
#define SCU_PLL_CTRL                    (*((volatile uint32_t *)(SCU_REG_BASE + 0x0C)))

#define SCU_CLKS_CLKSRCSEL_MASK                            (0x1U)
#define SCU_CLKS_CLKSRCSEL_SHIFT                           (0U)
#define SCU_CLKS_CLKSRCSEL(x)                              (((uint32_t)(((uint32_t)(x)) << SCU_CLKS_CLKSRCSEL_SHIFT)) & SCU_CLKS_CLKSRCSEL_MASK)
/* SCU_STS Register */
#define SCU_STS_PLL_LOCK_MASK                              (0x20U)
#define SCU_STS_PLL_LOCK_SHIFT                             (5U)
#define SCU_STS_PLL_LOCK(x)                                (((uint32_t)(((uint32_t)(x)) << SCU_STS_PLL_LOCK_SHIFT)) & SCU_STS_PLL_LOCK_MASK)
#define SCU_STS_CLKST_MASK                                 (0x1U)
#define SCU_STS_CLKST_SHIFT                                (0U)
#define SCU_STS_CLKST(x)                                   (((uint32_t)(((uint32_t)(x)) << SCU_STS_CLKST_SHIFT)) & SCU_STS_CLKST_MASK)
/* SCU_OSC_CTRL Register */
#define SCU_OSC_CTRL_OSC_DS_EN_MASK                        (0x2U)
#define SCU_OSC_CTRL_OSC_DS_EN_SHIFT                       (1U)
#define SCU_OSC_CTRL_OSC_DS_EN(x)                          (((uint32_t)(((uint32_t)(x)) << SCU_OSC_CTRL_OSC_DS_EN_SHIFT)) & SCU_OSC_CTRL_OSC_DS_EN_MASK)
/* SCU_PLL_CTRL Register */
#define SCU_PLL_CTRL_PLL_NDIV_MASK                         (0x1F00U)
#define SCU_PLL_CTRL_PLL_NDIV_SHIFT                        (8U)
#define SCU_PLL_CTRL_PLL_NDIV(x)                           (((uint32_t)(((uint32_t)(x)) << SCU_PLL_CTRL_PLL_NDIV_SHIFT)) & SCU_PLL_CTRL_PLL_NDIV_MASK)
#define SCU_PLL_CTRL_UNLOCK_RE_MASK                        (0x20U)
#define SCU_PLL_CTRL_UNLOCK_RE_SHIFT                       (5U)
#define SCU_PLL_CTRL_UNLOCK_RE(x)                          (((uint32_t)(((uint32_t)(x)) << SCU_PLL_CTRL_UNLOCK_RE_SHIFT)) & SCU_PLL_CTRL_UNLOCK_RE_MASK)
#define SCU_PLL_CTRL_PLL_OVRD_EN_MASK                      (0x4U)
#define SCU_PLL_CTRL_PLL_OVRD_EN_SHIFT                     (2U)
#define SCU_PLL_CTRL_PLL_OVRD_EN(x)                        (((uint32_t)(((uint32_t)(x)) << SCU_PLL_CTRL_PLL_OVRD_EN_SHIFT)) & SCU_PLL_CTRL_PLL_OVRD_EN_MASK)
#define SCU_PLL_CTRL_PLL_DS_EN_MASK                        (0x2U)
#define SCU_PLL_CTRL_PLL_DS_EN_SHIFT                       (1U)
#define SCU_PLL_CTRL_PLL_DS_EN(x)                          (((uint32_t)(((uint32_t)(x)) << SCU_PLL_CTRL_PLL_DS_EN_SHIFT)) & SCU_PLL_CTRL_PLL_DS_EN_MASK)
#define SCU_PLL_CTRL_PLL_EN_MASK                           (0x1U)
#define SCU_PLL_CTRL_PLL_EN_SHIFT                          (0U)
#define SCU_PLL_CTRL_PLL_EN(x)                             (((uint32_t)(((uint32_t)(x)) << SCU_PLL_CTRL_PLL_EN_SHIFT)) & SCU_PLL_CTRL_PLL_EN_MASK)


#define PAGE_SIZE_SHIFT                 (3)  // 8 bytes page size

/********************************************************************/
void strcopy(char *to, char* from);
static uint32_t _FlashExecuteCmd();

/*************************************************************************
 * Function Name: FlashInit
 * Parameters: base_of_flash: Flash Base Address
 *             image_size: the size of whole image, in bytes
 *             link_address: link addr
 *             flags: specifies original flags
 *
 * Return:  0 - Init Successful
 *          1 - Init Fail
 * Description: Init flash and build layout.
 *
 *************************************************************************/
#if USE_ARGC_ARGV
uint32_t FlashInit(void *base_of_flash, uint32_t image_size,
                   uint32_t link_address, uint32_t flags,
                   int argc, char const *argv[])
#else
uint32_t FlashInit(void *base_of_flash, uint32_t image_size,
                   uint32_t link_address, uint32_t flags)
#endif /* USE_ARGC_ARGV */
{
#if USE_ARGC_ARGV
    for(int i = 0; i < argc; i++)
    {
        /* if there is a partition arg */
        // TODO
    }
#else
    // TO DO
#endif

    /* disable watchdog */
    WDG_UNLOCK = 0xB631;
    WDG_UNLOCK = 0xC278;
    WDG_CONTROL &= ~(0x01UL);

    // 56MHz
    SCU_PLL_CTRL = SCU_PLL_CTRL_PLL_NDIV(28) | SCU_PLL_CTRL_PLL_EN(1);
    while((SCU_STS & SCU_STS_PLL_LOCK(1)) == 0);
    
    SCU_CLKS = SCU_CLKS_CLKSRCSEL(1);
    while((SCU_CLKS & SCU_STS_CLKST(1)) == 0);

    EFM_CTRL |= EFM_CTRL_PRESCALER(28);

    return _FlashExecuteCmd(EFM_NOP_CMD);
}


/*************************************************************************
 * Function Name: FlashWrite
 * Parameters: block_start: block base address
 *             offset_into_block: (uint32_t)block_start + offset_into_block
 *             count: the number of bytes to write
 *             buffer: ram buffer pointer
 * Return:  0 - Write Successful
 *          1 - Write Fail
 * Description. Writes data to Flash
 *************************************************************************/
uint32_t FlashWrite(void *block_start,
                    uint32_t offset_into_block,
                    uint32_t count,
                    char const *buffer)
{
    int32_t sz = count;
    int8_t i;
    volatile uint32_t *pSrc;
    volatile uint32_t *pDest;

    pSrc = (volatile uint32_t *)buffer;
    pDest = (volatile uint32_t *)((uint32_t)block_start + offset_into_block);
    
    EFM_CTRL |= EFM_CTRL_WE_MASK;

    do
    {
        for (i = 0; i < (EFM_INTERNAL_PAGE_SIZE/4); i++)
        {
            *pDest++ = *pSrc++;
        }
        
        if(_FlashExecuteCmd(EFM_PROG_64_CMD) != 0)
        {
            return RESULT_ERROR;
        }
        sz -= EFM_INTERNAL_PAGE_SIZE;
    } while(sz > 0);
    
    WDG_UNLOCK = 0xA518;
    WDG_UNLOCK = 0xD826;
    
    EFM_CTRL &= ~EFM_CTRL_WE_MASK;

    return RESULT_OK;
}


/*************************************************************************
 * Function Name: FlashErase
 * Parameters:  block_start: Block Address
 *              block_size: Block Size, maybe not used
 *
 * Return: 0
 *
 * Description: Erase block
 *************************************************************************/
uint32_t FlashErase(void *block_start,
                    uint32_t block_size)
{
    EFM_CTRL |= EFM_CTRL_WE_MASK;
    *(volatile uint32_t *)(block_start) = 0x12345678;
    if(_FlashExecuteCmd(EFM_SECTOR_ERASE_CMD) != 0)
    {
        return RESULT_ERROR;
    }
    EFM_CTRL &= ~EFM_CTRL_WE_MASK;

    return RESULT_OK;
}


/*************************************************************************
 * Function Name: _FlashExecuteCmd
 * Parameters: None
 *
 * Return: 0
 *
 * Description: wait flash to be ready for next command
**************************************************************************/
static uint32_t _FlashExecuteCmd(uint32_t code)
{
    EFM_STAT = EFM_STAT;
    if (code != EFM_NOP_CMD)
    {
        /* unlock flash */
        EFM_KEY = EFM_UNLOCK_KEY;
        /* execute command */
        EFM_CMD = (uint8_t)code;
    }
    while((EFM_STAT & EFM_STAT_CCIF_MASK) == 0);
    //return (EFM_STAT & EFM_ERROR_MASK);
    return 0;
}



OPTIONAL_CHECKSUM
uint32_t FlashChecksum(void const *begin, uint32_t count)
{
  return Crc16((uint8_t const *)begin, count);
}

OPTIONAL_SIGNOFF
uint32_t FlashSignoff(void)
{
  return 0;
}

void strcopy(char *to, char* from)
{
  while (*to++ = *from++) ;
}




